import { Logger } from '@oceanprotocol/lib'
import { useState } from 'react'
import { useOcean } from 'providers'
import { PriceOptions } from './PriceOptions'
import { TransactionReceipt } from 'web3-core'
import { getBestDataTokenPrice, getFirstPool } from 'utils/dtUtils'
import { Decimal } from 'decimal.js'

interface UsePricing {
  createPricing: (
    dataTokenAddress: string,
    priceOptions: PriceOptions
  ) => Promise<TransactionReceipt | string | null>
  buyDT: (
    dataTokenAddress: string,
    dtAmount: number | string
  ) => Promise<TransactionReceipt | null>
  sellDT: (
    dataTokenAddress: string,
    dtAmount: number | string
  ) => Promise<TransactionReceipt | null>
  mint: (tokenAddress: string, tokensToMint: string) => void
  pricingStep?: number
  pricingStepText?: string
  pricingError?: string
  pricingIsLoading: boolean
}

export const createPricingFeedback: { [key in number]: string } = {
  0: '1/4 Approving DT ...',
  1: '2/4 Approving Ocean ...',
  2: '3/4 Creating ....',
  3: '4/4 Pricing created'
}
export const buyDTFeedback: { [key in number]: string } = {
  0: '1/3 Approving OCEAN ...',
  1: '2/3 Buying DT ...',
  2: '3/3 DT Bought'
}
export const sellDTFeedback: { [key in number]: string } = {
  0: '1/3 Approving DT ...',
  1: '2/3 Selling DT ...',
  2: '3/3 DT sold'
}
function usePricing(): UsePricing {
  const { ocean, status, account, accountId, config } = useOcean()
  const [pricingIsLoading, setPricingIsLoading] = useState(false)
  const [pricingStep, setPricingStep] = useState<number | undefined>()
  const [pricingStepText, setPricingStepText] = useState<string | undefined>()
  const [pricingError, setPricingError] = useState<string | undefined>()
  const [dtSymbol, setDtSymbol] = useState<string>()

  function setStepCreatePricing(index?: number) {
    setPricingStep(index)
    let message
    if (index) {
      if (dtSymbol)
        message = createPricingFeedback[index].replace(/DT/g, dtSymbol)
      else message = createPricingFeedback[index]
      setPricingStepText(message)
    }
  }

  function setStepBuyDT(index?: number) {
    setPricingStep(index)
    let message
    if (index) {
      if (dtSymbol) message = buyDTFeedback[index].replace(/DT/g, dtSymbol)
      else message = buyDTFeedback[index]
      setPricingStepText(message)
    }
  }
  function setStepSellDT(index?: number) {
    setPricingStep(index)
    let message
    if (index) {
      if (dtSymbol) message = sellDTFeedback[index].replace(/DT/g, dtSymbol)
      else message = sellDTFeedback[index]
      setPricingStepText(message)
    }
  }

  async function mint(tokenAddress: string, tokensToMint: string) {
    Logger.log('mint function', tokenAddress, accountId)
    await ocean.datatokens.mint(tokenAddress, accountId, tokensToMint)
  }
  async function buyDT(
    dataTokenAddress: string,
    dtAmount: number | string
  ): Promise<TransactionReceipt | null> {
    if (!ocean || !account || !accountId) return null

    try {
      setDtSymbol(await ocean.datatokens.getSymbol(dataTokenAddress))
      setPricingIsLoading(true)
      setPricingError(undefined)
      setStepBuyDT(0)
      const bestPrice = await getBestDataTokenPrice(ocean, dataTokenAddress)
      switch (bestPrice?.type) {
        case 'pool': {
          const price = new Decimal(bestPrice.value).times(1.05).toString()
          const maxPrice = new Decimal(bestPrice.value).times(2).toString()
          setStepBuyDT(1)
          Logger.log(
            'Buying token from pool',
            bestPrice,
            account.getId(),
            price
          )
          const buyResponse = await ocean.pool.buyDT(
            account.getId(),
            bestPrice.address,
            String(dtAmount),
            price,
            maxPrice
          )
          setStepBuyDT(2)
          Logger.log('DT buy response', buyResponse)
          return buyResponse
        }
        case 'exchange': {
          if (!config.oceanTokenAddress) {
            Logger.error(`'oceanTokenAddress' not set in config`)
            return null
          }
          if (!config.fixedRateExchangeAddress) {
            Logger.error(`'fixedRateExchangeAddress' not set in config`)
            return null
          }
          Logger.log('Buying token from exchange', bestPrice, account.getId())
          await ocean.datatokens.approve(
            config.oceanTokenAddress,
            config.fixedRateExchangeAddress,
            bestPrice.value.toString(),
            account.getId()
          )
          setStepBuyDT(1)
          const exchange = await ocean.fixedRateExchange.buyDT(
            bestPrice.address,
            String(dtAmount),
            account.getId()
          )
          setStepBuyDT(2)
          Logger.log('DT exchange buy response', exchange)
          return exchange
        }
      }
    } catch (error) {
      setPricingError(error.message)
      Logger.error(error)
    } finally {
      setStepBuyDT(undefined)
      setPricingStepText(undefined)
      setPricingIsLoading(false)
    }
    return null
  }

  async function sellDT(
    dataTokenAddress: string,
    dtAmount: number | string
  ): Promise<TransactionReceipt | null> {
    if (!ocean || !account || !accountId) return null
    if (!config.oceanTokenAddress) {
      Logger.error(`'oceanTokenAddress' not set in config`)
      return null
    }
    try {
      setDtSymbol(await ocean.datatokens.getSymbol(dataTokenAddress))
      setPricingIsLoading(true)
      setPricingError(undefined)
      setStepSellDT(0)
      const pool = await getFirstPool(ocean, dataTokenAddress)
      if (!pool || pool.price === 0) return null
      const price = new Decimal(pool.price).times(0.95).toString()
      setStepSellDT(1)
      Logger.log('Selling token to pool', pool, account.getId(), price)
      const sellResponse = await ocean.pool.sellDT(
        account.getId(),
        pool.address,
        String(dtAmount),
        price
      )
      setStepSellDT(2)
      Logger.log('DT sell response', sellResponse)
      return sellResponse
    } catch (error) {
      setPricingError(error.message)
      Logger.error(error)
    } finally {
      setStepSellDT(undefined)
      setPricingStepText(undefined)
      setPricingIsLoading(false)
    }
    return null
  }

  async function createPricing(
    dataTokenAddress: string,
    priceOptions: PriceOptions
  ): Promise<TransactionReceipt | string | null> {
    if (!ocean || !account || !accountId) return null

    let response = null
    try {
      setPricingIsLoading(true)
      setPricingError(undefined)
      setDtSymbol(await ocean.datatokens.getSymbol(dataTokenAddress))
      setStepCreatePricing(0)
      switch (priceOptions.type) {
        case 'dynamic': {
          setStepCreatePricing(2)
          response = await ocean.pool.createDTPool(
            accountId,
            dataTokenAddress,
            priceOptions.dtAmount.toString(),
            priceOptions.weightOnDataToken,
            priceOptions.swapFee
          )
          setStepCreatePricing(3)
          return response
        }
        case 'fixed': {
          if (!config.fixedRateExchangeAddress) {
            Logger.error(`'fixedRateExchangeAddress' not set in ccnfig.`)
            return null
          }
          setStepCreatePricing(2)
          response = await ocean.fixedRateExchange.create(
            dataTokenAddress,
            priceOptions.price.toString(),
            accountId
          )
          setStepCreatePricing(1)
          await ocean.datatokens.approve(
            dataTokenAddress,
            config.fixedRateExchangeAddress,
            String(priceOptions.dtAmount),
            accountId
          )
          setStepCreatePricing(3)
          return response
        }
      }
    } catch (error) {
      setPricingError(error.message)
      Logger.error(error)
    } finally {
      setPricingStep(undefined)
      setPricingStepText(undefined)
      setPricingIsLoading(false)
    }
    return null
  }

  return {
    createPricing,
    buyDT,
    sellDT,
    mint,
    pricingStep,
    pricingStepText,
    pricingIsLoading,
    pricingError
  }
}

export { usePricing, UsePricing }
export default usePricing
